$(function() {
	var Engine = window.Engine = Stella()
								 .include('Input, Sprites, Scenes') //, DOM') // Remove DOM for Canvas Game
								 //.domOnly()								//  Remove domOnly for Canvas Game
								 .setup("stella", { "maximize": false });
							
	Engine.input.keyboardControls();
	Engine.input.touchControls({
		controls: [ ['left', '<'], [], [], [], ['right', '>'] ]
	});
	
	Engine.Paddle = Engine.Sprite.extend({
		init: function() {
			this._super({
				sheet: 'paddle',
				speed: 200,
				x:0
			});
			
			this.p.x = Engine.width/2 - this.p.w /2;
			this.p.y = Engine.height - this.p.h;
			if(Engine.input.keypad.size) {
				this.p.y -= Engine.input.keypad.size + this.p.h;
			}
		},
		
		step: function(dt) {
			if(Engine.inputs['left']) {
				this.p.x -= dt * this.p.speed;
			} else if(Engine.inputs['right']) {
				this.p.x += dt * this.p.speed;
			}
			if(this.p.x < 0) {
				this.p.x = 0;
			} else if (this.p.x > Engine.width - this.p.w) {
				this.p.x = Engine.width - this.p.w;
			}
			this._super(dt);
		}
		
	});
	
	Engine.Ball = Engine.Sprite.extend({
		init: function() {
			this._super({
				sheet: 'ball',
				speed: 200,
				dx: 1,
				dy: -1
			});
			this.p.y = Engine.height / 2 - this.p.h;
			this.p.x = Engine.width /2 + this.p.w /2;
		},
		
		step: function(dt) {
			var p = this.p;
			
			// collision check
			var hit = Engine.stage().collide(this);
			if(hit) {
				if(hit instanceof Engine.Paddle) {
					p.dy = -1;
				} else {
					hit.trigger('collision', this);
				}
			}
			
			p.x += p.dx * p.speed * dt;
			p.y += p.dy * p.speed * dt;
			
			// check x collision
			if(p.x< 0) {
				p.x = 0;
				p.dx = 1;
			} else if(p.x > Engine.width - p.w) {
				p.dx = -1;
				p.x = Engine.width - p.w;
			}
			
			// check y collision
			if(p.y < 0) {
				p.y = 0;
				p.dy = 1;
			} else if(p.y > Engine.height) {
				Engine.stageScene('game');
			}
			this._super(dt);
		}
	});
	
	// blocks class
	Engine.Block = Engine.Sprite.extend({
		init: function(props) {
			this._super(_(props).extend({ sheet: 'block' }));
			this.bind('collision', function(ball) {
				this.destroy();
				ball.p.dy *= -1;
				
				Engine.stage().trigger('removeBlock');
			});
		},
	});
	
	Engine.load(['blockbreak.png', 'blockbreak.json'], function() {
		Engine.compileSheets('blockbreak.png', 'blockbreak.json');
		
		Engine.scene('game', new Engine.Scene(function(stage) {
			stage.insert(new Engine.Paddle());
			stage.insert(new Engine.Ball());
			
			var blockCount = 0;
			for(var x=0;x<6;x++) {
				for(var y=0; y<5; y++) {
					stage.insert(new Engine.Block({ x: x*50+10, y: y*30+10 }));
					blockCount++;
				}
			}
			
			stage.bind('removeBlock', function() {
				blockCount--;
				if(blockCount == 0) {
					Engine.stageScene('game');
				}
			});
		}));
		
		Engine.stageScene('game');
	});
});