Stella.Scenes = function(Engine) {

  Engine.scenes = {};
  Engine.stages = [];

  Engine.Scene = Class.extend({
    init: function(sceneFunc,opts) {
      this.opts = opts || {};
      this.sceneFunc = sceneFunc;
    }
  });

  // Set up or return a new scene
  Engine.scene = function(name,sceneObj) {
    if(!sceneObj) {
      return Engine.scenes[name];
    } else {
      Engine.scenes[name] = sceneObj;
      return sceneObj;
    }
  };


  Engine.overlap = function(o1,o2) {
    return !((o1.p.y+o1.p.h-1<o2.p.y) || (o1.p.y>o2.p.y+o2.p.h-1) ||
             (o1.p.x+o1.p.w-1<o2.p.x) || (o1.p.x>o2.p.x+o2.p.w-1));
  };

  Engine.Stage = Engine.GameObject.extend({
    // Should know whether or not the stage is paused
    defaults: {
      sort: false
    },

    init: function(scene) {
      this.scene = scene;
      this.items = [];
      this.index = {};
      this.removeList = [];
      if(scene)  { 
        this.options = _(this.defaults).clone();
        _(this.options).extend(scene.opts);
        scene.sceneFunc(this);
      }
      if(this.options.sort && !_.isFunction(this.options.sort)) {
          this.options.sort = function(a,b) { return a.p.z - b.p.z; };
      }
    },

    each: function(callback) {
      for(var i=0,len=this.items.length;i<len;i++) {
        callback.call(this.items[i],arguments[1],arguments[2]);
      }
    },

    eachInvoke: function(funcName) {
      for(var i=0,len=this.items.length;i<len;i++) {              
        this.items[i][funcName].call(
          this.items[i],arguments[1],arguments[2]
        );
      }
    },

    detect: function(func) {
      for(var i = 0,val=null, len=this.items.length; i < len; i++) {
        if(func.call(this.items[i],arguments[1],arguments[2])) {
          return this.items[i];
        }
      }
      return false;
    },

    insert: function(itm) {
      this.items.push(itm);
      itm.parent = this;
      if(itm.p) {
        this.index[itm.p.id] = itm;
      }
      this.trigger('inserted',itm);
      itm.trigger('inserted',this);
      return itm;
    },

    remove: function(itm) {
      this.removeList.push(itm);
    },

    forceRemove: function(itm) {
      var idx = _(this.items).indexOf(itm);
      if(idx != -1) { 
        this.items.splice(idx,1);
        if(itm.destroy) itm.destroy();
        if(itm.p.id) {
          delete this.index[itm.p.id];
        }
        this.trigger('removed',itm);
      }
    },

    pause: function() {
      this.paused = true;
    },

    unpause: function() {
      this.paused = false;
    },

    _hitTest: function(obj,type) {
      if(obj != this) {
        var col = (!type || this.p.type & type) && Engine.overlap(obj,this);
        return col ? this : false;
      }
    },

    collide: function(obj,type) {
      return this.detect(this._hitTest,obj,type);
    },

    step:function(dt) {
      if(this.paused) { return false; }

      this.trigger("prestep",dt);
      this.eachInvoke("step",dt);
      this.trigger("step",dt);

      if(this.removeList.length > 0) {
        for(var i=0,len=this.removeList.length;i<len;i++) {
          this.forceRemove(this.removeList[i]);
        }
        this.removeList.length = 0;
      }
    },

    draw: function(ctx) {
      if(this.options.sort) {
        this.items.sort(this.options.sort);
      }
      this.trigger("predraw",ctx);
      this.eachInvoke("draw",ctx);
      this.trigger("draw",ctx);
    }
  });

  Engine.activeStage = 0;

  Engine.stage = function(num) {
    // Use activeStage is num is undefined
    num = (num === void 0) ? Engine.activeStage : num;
    return Engine.stages[num];
  };

  Engine.stageScene = function(scene,num,stageClass) {
    stageClass = stageClass || Engine.Stage;
    if(_(scene).isString()) {
      scene = Engine.scene(scene);
    }

    num = num || 0;

    if(Engine.stages[num]) {
      Engine.stages[num].destroy();
    }

    Engine.stages[num] = new stageClass(scene);

    if(!Engine.loop) {
      Engine.gameLoop(Engine.stageGameLoop);
    }
  };

  Engine.stageGameLoop = function(dt) {
    if(Engine.ctx) { Engine.clear(); }

    for(var i =0,len=Engine.stages.length;i<len;i++) {
      Engine.activeStage = i;
      var stage = Engine.stage();
      if(stage) {
        stage.step(dt);
        stage.draw(Engine.ctx);
      }
    }

    Engine.activeStage = 0;

    if(Engine.input && Engine.ctx) { Engine.input.drawCanvas(Engine.ctx); }
  };

  Engine.clearStage = function(num) {
    if(Engine.stages[num]) { 
      Engine.stages[num].destroy(); 
      Engine.stages[num] = null;
    }
  };

  Engine.clearStages = function() {
    for(var i=0,len=Engine.stages.length;i<len;i++) {
      if(Engine.stages[i]) { Engine.stages[i].destroy(); }
    }
    Engine.stages.length = 0;
  };


};

