/* simple javascript inheritance
*/

// inspired by base2 and Prototype
(function() {
	var initializing = false,
		fnTest = /xyz/.test(function() {xyz;}) ? /\b_super\b/ : /.*/;
		
		// the base class implementation (does nothing)
		this.Class = function() {};
		
		// create a new class that inherits from this class
		Class.extend = function(prop) {
			var _super = this.prototype;
			
			// instantiate  a base class (but only create the instance, don't run init constructor)
			initializing = true;
			var prototype = new this();
			initializing = false;
			
			// copy the properties over onto the new prototype
			for (var name in prop) {
				// check if we're overwritting an existing function
				prototype[name] = typeof prop[name] == "function" && typeof _super[name] == "function" && 
				fnTest.test(prop[name]) ?
					
					(function(name, fn) {
						return function() {
							var tmp = this._super;
							
							// add new ._super() method that is the same method 
							// but on the super-class
							this._super = _super[name];
							
							// the method only need to be bound temporarily, so we 
							// remove it when we're don executing
							var ret = fn.apply(this, arguments);
							this._super = tmp;
							
							return ret;
						};
					})(name, prop[name]) : prop[name];
			}
			
			// the dummy class constructor
			function Class() {
				// all construction is actually done in the init method
				if (!initializing && this.init)
					this.init.apply(this, arguments);
			}
			
			// populate our constructed prototype object
			Class.prototype = prototype;
			
			// enforce the constructor to be what we expect
			Class.prototype.constructor = Class;
			// and make this class extendable
			Class.extend = arguments.callee;
			
			return Class;
		};
})();